package gp;

import simplerace.Evolvable;

import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Random;

/**
 * Created by IntelliJ IDEA.
 * User: julian
 * Date: Feb 8, 2007
 * Time: 1:25:31 AM
 */
public class DoubleGP implements Evolvable, Base {
    
    private double[] inputs;
    private Node[] roots;
    private Node[] ADFs;
    private double[] cachedADFValues;
    private Random random = new Random ();
    private int maximumDepth;
    private double probabilityOfTerminal = 0.7; // indirectly affects the average depth of the tree

    private final Class[] terminals = {Constant.class, Input.class, ADFCall.class};
    private final Class[] nonTerminals = {Arithmetic.class, Trigonometric.class, IfThenElse.class};

    private final boolean cacheADFValues = true;

    public DoubleGP(int numberOfInputs, int numberOfOutputs, int maximumDepth, int numberOfADFs) {
        this.maximumDepth = maximumDepth;
        this.inputs = new double[numberOfInputs];
        roots = new Node[numberOfOutputs];
        ADFs = new Node[numberOfADFs];
        cachedADFValues = new double[ADFs.length];
        for (int i = 0; i < roots.length; i++) {
            roots[i] = newNode (maximumDepth, i);
        }
        for (int i = 0; i < ADFs.length; i++) {
            ADFs[i] = newNode (maximumDepth, roots.length + i);
        }
        reset ();
    }

    public DoubleGP(int numberOfInputs, Node[] roots, int maximumDepth, Node[] ADFs) {
        this.maximumDepth = maximumDepth;
        this.inputs = new double[numberOfInputs];
        this.roots = roots;
        this.ADFs = ADFs;
        cachedADFValues = new double[ADFs.length];
        reset ();
    }

    public Evolvable getNewInstance() {
        return new DoubleGP (inputs.length, roots.length, maximumDepth, ADFs.length);
    }

    public DoubleGP copy() {
        Node[] rootsCopy = new Node[roots.length];
        for (int i = 0; i < roots.length; i++) {
            rootsCopy[i] = roots[i].copy ();
        }
        Node[] adfsCopy = new Node[ADFs.length];
        for (int i = 0; i < ADFs.length; i++) {
            adfsCopy[i] = ADFs[i].copy ();
        }
        return new DoubleGP (inputs.length, rootsCopy, maximumDepth, adfsCopy);
    }

    public void reset() {  
        for (int i = 0; i < cachedADFValues.length; i++) {
            cachedADFValues[i] = Double.NaN;
        }
    }

    public double[] approximate(double[] doubles) {
        if (inputs.length != doubles.length) {
            throw new ArrayIndexOutOfBoundsException ("Numbers of inputs don't match. Inputs in FA "
                    + inputs.length + ", but the task gives me " + doubles.length);

        }
        System.arraycopy (doubles, 0, inputs, 0, doubles.length);
        double[] outputs = new double[roots.length];
        for (int i = 0; i < outputs.length; i++) {
            outputs[i] = roots[i].value (this);
        }
        return outputs;
    }

    public void mutate() {
        // one-point macro mutation
        for (int i = 0; i < roots.length; i++) {
            roots[i] = onePointMacroMutate (roots[i], i);
        }
        for (int i = 0; i < ADFs.length; i++) {
            ADFs[i] = onePointMacroMutate (ADFs[i], roots.length + i);
        }
    }

    private Node onePointMacroMutate(Node node, int treeId) {
        // create a list of all nodes
        ArrayList<Node> nodes = new ArrayList<Node> (node.size ());
        node.listNodes (nodes);
        if (Math.random () < (1.0 / nodes.size ())) {
            // with a small probability, replace the root
            node = newNode (maximumDepth, treeId);
        } else {
            // otherwise select one node at random and mutate
            Node victim = nodes.get ((int) (Math.random () * nodes.size ()));
            victim.mutate (this);
        }
        return node;
    }

    public double gaussian() {
        return random.nextGaussian ();
    }

    public int numberOfInputs() {
        return inputs.length;
    }

    public double valueOfInput(int which) {
        return inputs[which];
    }

    public double valueOfADF(int which) {
        if (!cacheADFValues) {
            return ADFs[which].value (this);
        }
        else {
            if (cachedADFValues[which] == Double.NaN) {
                cachedADFValues[which] = ADFs[which].value (this);
            }
            return cachedADFValues[which];
        }
    }

    public int getADF (int treeId) {
        if ((treeId - roots.length) >= ADFs.length  || ADFs.length == 0) {
            throw new RuntimeException ("Trying to get an ADF, but caller's treeId is "
                    + treeId + " and we have " + ADFs.length + " ADFs");
        } else {
            int firstEligible = (Math.max (0, treeId - roots.length + 1));
            int numberOfPossible = ADFs.length - firstEligible;
            return firstEligible + (int) (Math.random () * numberOfPossible);
        }
    }

    public Node newNode(int maximumDepth, int treeId) {
        if (maximumDepth == 0 || Math.random () > probabilityOfTerminal) {
            Class<Node> terminalClass;
            do {
                terminalClass = terminals[(int) (Math.random () * terminals.length)];
            }
            // don't create any adfcalls for the last adf
            while (treeId == (roots.length + ADFs.length -1) &&
                    terminalClass == (Class<? extends Node>) ADFCall.class);
            try {
                Constructor constructor = terminalClass.getConstructor (Base.class, Integer.class);
                //System.out.println (constructor);
                return (Node) constructor.newInstance (this, treeId);
            }
            catch (Exception e) {
                e.printStackTrace ();
                System.exit (-1);
                return null;
            }
        } else {
            Class<Node> nonTerminalClass = nonTerminals[(int) (Math.random () * nonTerminals.length)];
            try {
                Constructor constructor = nonTerminalClass.getConstructor (Base.class, Integer.class, Integer.class);
                //System.out.println (constructor);
                return (Node) constructor.newInstance (this, new Integer (maximumDepth), new Integer (treeId));
            }
            catch (Exception e) {
                e.printStackTrace ();
                System.exit (-1);
                return null;
            }
        }
    }

    public String toString () {
        StringBuffer sb = new StringBuffer ("DoubleGP, " + roots.length + " roots, " + ADFs.length + " ADFs\n");
       /* for (int i = 0; i < roots.length; i++) {
            sb.append ("   root " + i + ": " + roots[i] + "\n");
            sb.append ("   root " + i + " size: " + roots[i].size () + "\n");
        }
        for (int i = 0; i < ADFs.length; i++) {
            sb.append ("    ADF " + i + ": " + ADFs[i] + "\n");
            sb.append ("    ADF " + i + " size: " + ADFs[i].size () + "\n");
        }  */
        return sb.toString ();
    }
}
