package simplerace;


import java.util.Random;

/**
 * Created by IntelliJ IDEA.
 * User: jtogel
 * Date: 17-Oct-2006
 * Time: 16:29:09
 */
public class BasicTrack implements Track, Constants {

    static final int width = 400;
    static final int height = 400;
    static final int diagonal = (int) Math.sqrt (width * width + height * height);

    static final int nPoints = 3;
    private Vector2d[] waypoints;
    private Vehicle firstVehicle = null;
    private Vehicle secondVehicle = null;
    private SensorModel firstSensorModel = null;
    private SensorModel secondSensorModel = null;
    private int current;
    private int firstWaypoints = 0;
    private int secondWaypoints = 0;
    private Random random = new Random ();
    private boolean firstVehicleJustPassedAWaypoint = false;
    private boolean secondVehicleJustPassedAWaypoint = false;

    public BasicTrack () {
        waypoints = new Vector2d[nPoints];
        reset ();
    }

    public BasicTrack (Vector2d[] waypoints, int current, int firstWaypoints, int secondWaypoints) {
        this.waypoints = waypoints;
        this.current = current;
        this.firstWaypoints = firstWaypoints;
        this.secondWaypoints = secondWaypoints;
    }

    public void reset () {
        firstWaypoints = 0;
        secondWaypoints = 0;
        for (int i = 0; i < waypoints.length; i++) {
            waypoints[i] = randomPoint ();
        }
    }

    public void setFirstVehicle (Vehicle vehicle, boolean reposition) {
        this.firstVehicle = vehicle;
        if (reposition) {
            firstVehicle.setPosition (width / 2, height / 2 - 20);
            firstVehicle.setOrientation (0);
        }
        firstSensorModel = new IntegratedSensorModel (firstVehicle);
    }

    public void setSecondVehicle (Vehicle vehicle, boolean reposition) {
        this.secondVehicle = vehicle;
        if (vehicle != null) {
            if (reposition) {
                secondVehicle.setPosition (width / 2, height / 2 + 20);
                secondVehicle.setOrientation (Math.PI);
            }
            secondSensorModel = new IntegratedSensorModel (secondVehicle);
        }
        else {
            secondSensorModel = null;
        }
    }

    public void update () {
        firstVehicleJustPassedAWaypoint = false;
        secondVehicleJustPassedAWaypoint = false;
        if (random.nextDouble () < 0.5) {
            // check first vehicle first, and second vehicle afterward
            if (firstVehicle != null &&
                    (firstVehicle.getPosition ().dist (waypoints[current]) <= waypointIntersectionRadius)) {
                nextPoint ();
                firstWaypoints++;
                firstVehicleJustPassedAWaypoint = true;
            }
            if (secondVehicle != null &&
                    (secondVehicle.getPosition ().dist (waypoints[current]) <= waypointIntersectionRadius)) {
                nextPoint ();
                secondWaypoints++;
                secondVehicleJustPassedAWaypoint = true;
            }
        }
        else {
            // check second vehicle first, and first vehicle afterwards
            if (secondVehicle != null &&
                    (secondVehicle.getPosition ().dist (waypoints[current]) <= waypointIntersectionRadius)) {
                nextPoint ();
                secondWaypoints++;
                secondVehicleJustPassedAWaypoint = true;
            }
            if (firstVehicle != null &&
                    (firstVehicle.getPosition ().dist (waypoints[current]) <= waypointIntersectionRadius)) {
                nextPoint ();
                firstWaypoints++;
                firstVehicleJustPassedAWaypoint = true;
            }
        }
    }

    public SensorModel getFirstSensorModel () {
        return firstSensorModel;
    }

    public SensorModel getSecondSensorModel () {
        return secondSensorModel;
    }

    private void nextPoint() {
        waypoints[current] = randomPoint();
        current = (current + 1) % nPoints;
    }

    public Vector2d randomPoint() {
        return new Vector2d (30 + random.nextDouble () * (width - 60),
                30 + random.nextDouble () * (height - 60));
    }

    public Vector2d[] getAllWaypointsInOrder () {
        Vector2d[] inOrder = new Vector2d[waypoints.length];
        for (int i = 0; i < inOrder.length; i++) {
            inOrder[i] = new Vector2d (waypoints[(current + i) % nPoints]);
        }
        return inOrder;
    }

    public double getFirstFitness () {
        return firstWaypoints;
    }

    public double getSecondFitness () {
        return secondWaypoints;
    }

    public int getWidth () {
        return width;
    }

    public int getHeight () {
        return height;
    }

    public Track copy () {
        // returns a complete copy  of the track, including all state variables EXCEPT the vehicles
        Vector2d[] waypointCopies = new Vector2d[waypoints.length];
        for (int i = 0; i < waypoints.length; i++) {
            waypointCopies[i] = new Vector2d(waypoints[i]);
        }
        return new BasicTrack (waypointCopies, current, firstWaypoints, secondWaypoints);
    }

    private class IntegratedSensorModel implements SensorModel {

        private final Vehicle vehicle;

        public IntegratedSensorModel (Vehicle vehicle) {
            this.vehicle = vehicle;
            if (this.vehicle == null) {
                throw new RuntimeException ("This sensormodel is initialized with a null vehicle");
            }
        }

        private Vehicle getOtherVehicle () {
            if (firstVehicle == null || secondVehicle == null) {
                return null;
            }
            if (this.vehicle == firstVehicle) {
                return secondVehicle;
            }
            else if (this.vehicle == secondVehicle) {
                return firstVehicle;
            }
            else throw new RuntimeException ("Vehicle does not exist");
        }

        public double getSpeed () {
            Vector2d velocity = vehicle.getVelocity ();
            double velocityMagnitude = velocity.mag ();
            double directionOfMovement = Math.atan2 (velocity.y, velocity.x);
            double directionDifference = directionOfMovement -
                vehicle.getOrientation ();
            return Math.cos (directionDifference) * velocityMagnitude;
        }

        public double getAngleToNextWaypoint () {
            Vector2d nextWp = waypoints[current];
            Vector2d position = vehicle.getPosition ();
            double xDiff = nextWp.x - position.x;
            double yDiff = nextWp.y - position.y;
            double angle = Math.atan2 (yDiff, xDiff);
            angle = vehicle.getOrientation () - angle;
            return Utils.correctAngle (angle);
        }

        public double getDistanceToNextWaypoint () {
            Vector2d nextWaypoint = waypoints[current];
            return vehicle.getPosition ().dist (nextWaypoint) / diagonal;
        }

        public double getAngleToNextNextWaypoint () {
            Vector2d nextNextWp = waypoints[(current + 1) % nPoints];
            Vector2d position = vehicle.getPosition ();
            double xDiff = nextNextWp.x - position.x;
            double yDiff = nextNextWp.y - position.y;
            double angle = Math.atan2 (yDiff, xDiff);
            angle = vehicle.getOrientation () - angle;
            return Utils.correctAngle (angle);
        }

        public double getDistanceToNextNextWaypoint () {
            Vector2d nextNextWaypoint = waypoints[(current + 1) % nPoints];
            return vehicle.getPosition ().dist (nextNextWaypoint) / diagonal;
        }

        public double getAngleToOtherVehicle () {
            Vehicle otherVehicle = getOtherVehicle ();
            if (otherVehicle == null) {
                return 0;
            }
            else {
                Vector2d otherPosition = otherVehicle.getPosition ();
                Vector2d thisPosition = vehicle.getPosition ();
                double xDiff = otherPosition.x - thisPosition.x;
                double yDiff = otherPosition.y - thisPosition.y;
                double angle = Math.atan2 (yDiff, xDiff);
                angle = vehicle.getOrientation () - angle;
                return Utils.correctAngle (angle);
            }
        }

        public double getDistanceToOtherVehicle () {
            Vehicle otherVehicle = getOtherVehicle ();
            if (otherVehicle == null) {
                return 1;
            }
            else {
                return (vehicle.getPosition ().dist (otherVehicle.getPosition ())) / diagonal;
            }
        }

        public boolean otherVehicleIsPresent () {
            return (getOtherVehicle () != null);
        }

        public boolean justPassedAWaypoint() {
            return (vehicle == firstVehicle) ? firstVehicleJustPassedAWaypoint : secondVehicleJustPassedAWaypoint;
        }

        public boolean otherVehicleJustPassedAWaypoint() {
            return (vehicle == secondVehicle) ? firstVehicleJustPassedAWaypoint : secondVehicleJustPassedAWaypoint;
        }

        public boolean justCollidedWithOtherVehicle() {
            return vehicle.justCollided ();
        }

        public Vector2d getPosition () {
            return vehicle.getPosition ().copy ();
        }

        public Vector2d getVelocity () {
            return vehicle.getVelocity ().copy ();
        }

        public double getOrientation () {
            return vehicle.getOrientation ();
        }

        public double getAngularVelocity() {
            return vehicle.getAngularVelocity ();
        }

        public Vector2d getOtherVehiclePosition () {
            Vehicle otherVehicle = getOtherVehicle ();
            if (otherVehicle == null) {
                return new Vector2d (0, 0);
            }
            else {
                return otherVehicle.getPosition ().copy ();
            }
        }

        public Vector2d getOtherVehicleVelocity () {
            Vehicle otherVehicle = getOtherVehicle ();
            if (otherVehicle == null) {
                return new Vector2d (0, 0);
            }
            else {
                return otherVehicle.getVelocity ().copy ();
            }
        }

        public double getOtherVehicleOrientation () {
            Vehicle otherVehicle = getOtherVehicle ();
            if (otherVehicle == null) {
                return 0;
            }
            else {
                return otherVehicle.getOrientation ();
            }
        }

        public double getOtherVehicleAngularVelocity() {
            return getOtherVehicle ().getAngularVelocity ();
        }

        public Vector2d getNextWaypointPosition () {
            return waypoints[current].copy ();
        }

        public Vector2d getNextNextWaypointPosition () {
            return waypoints[(current + 1) % nPoints].copy ();
        }

        public Simulation getSimulation() {
            Vehicle firstVehicleCopy = null;
            if (firstVehicle != null) {
                firstVehicleCopy = firstVehicle.copy ();
            }
            Vehicle secondVehicleCopy = null;
            if (secondVehicle != null) {
                secondVehicleCopy = secondVehicle.copy ();
            }
            // make sure that the controller always controls the red car in the returned simulation
            if (vehicle == firstVehicle) {
                return new BasicSimulator (copy (), firstVehicleCopy, secondVehicleCopy);
            }
            else {
                return new BasicSimulator (copy (), secondVehicleCopy, firstVehicleCopy);
            }
        }
    }
}
