package simplerace;

import java.awt.*;

/**
 * Created by IntelliJ IDEA.
 * User: jtogel
 * Date: 07-Oct-2006
 * Time: 22:51:45
 */
public class Car implements Vehicle, Constants {

    // constants
    final double mass = 10;
    final double constantDrag = 0.2;
    final double motorForceForward = 4;
    final double motorForceBackward = 2;
    final double turningRadius = 25; // pixels. effective at low speeds
    final double maxLateralTyreTraction = 2;
    final double minimumSlipAngleToCareAbout = Math.PI / 16; // radians. for numerical stability

    // state of the car (complete specification)
    private Vector2d velocity = new Vector2d ();
    private Vector2d position = new Vector2d ();
    private double orientation; // where the car is pointing; need not be where it is going
    private double angularVelocity;

    // convenience variables - dependent on the state variables
    private double speed; // speed in the direction the car is oriented
    //private double velocityMagnitude; // speed in the direction the car is moving
    private double directionOfMovement;
    private int[][] currentCorners = null;

    // for car-to-car collisions
    private boolean carCollisionThisTimeStep = false;
    private boolean carCollisionPreviousTimeStep = false;
    private Polygon polygon = null;

    public Car () {}

    public Car (Vector2d velocity, Vector2d position, double orientation, double angularVelocity,
                boolean carCollisionThisTimeStep, boolean carCollisionPreviousTimeStep){
        this.velocity = velocity;
        this.position = position;
        this.orientation = orientation;
        this.angularVelocity = angularVelocity;
        this.carCollisionThisTimeStep = carCollisionThisTimeStep;
        this.carCollisionPreviousTimeStep = carCollisionPreviousTimeStep;
    }

    public boolean update (int command) {
        carCollisionPreviousTimeStep = carCollisionThisTimeStep;
        carCollisionThisTimeStep = false;
        orientation = correctAngle (orientation);
        calculateSpeeds ();
        // the total force is first calculated, and then added to the velocity of the car
        Vector2d totalForce = new Vector2d ();
        // first, add force from drag
        totalForce.set (velocity);
        totalForce.mul (-constantDrag);       
        // add force from driving
        totalForce.add (getDrivingForce (command));
        // update angular velocity
        angularVelocity += (getAVFromSteering (command) / 2);
        // rotate car according to angular velocity
        orientation += angularVelocity;
        // Apply force from wheels - rotate velocity
        double slipAngle = directionOfMovement - (speed >= 0 ? orientation : orientation + Math.PI);
        slipAngle = correctAngle (slipAngle);
        if (Math.abs (slipAngle) > minimumSlipAngleToCareAbout) {
            totalForce.add (getLateralForce (slipAngle));
        }
        // Change velocity according to totalForce
        Vector2d acceleration = new Vector2d (totalForce.x / mass, totalForce.y / mass);
        velocity.add (acceleration);
        // Alright... now, change position of the car according to the velocity
        position = new Vector2d (position.x + (velocity.x),
                position.y + (velocity.y));
        currentCorners = null;
        polygon = null;
        return false;
    }

    public Vector2d getPosition () {
        return new Vector2d (position.x, position.y);
    }

    public Vector2d getVelocity () {
        return new Vector2d (velocity.x, velocity.y);
    }

    public double getOrientation () {
        return orientation;
    }

    public double getAngularVelocity() {
        return angularVelocity;
    }

    public void setPosition (double x, double y) {
        this.position.set (x, y);
    }

    public void setVelocity (double x, double y) {
        this.velocity.set (x, y);
    }

    public void setOrientation (double orientation) {
        this.orientation = orientation;
    }

    public void setAngularVelocity(double angularVelocity) {
        this.angularVelocity = angularVelocity;
    }

    public Vehicle copy() {
        Vector2d velocityCopy = new Vector2d (velocity);
        Vector2d positionCopy = new Vector2d (position);
        return new Car (velocityCopy, positionCopy, orientation, angularVelocity,
                carCollisionThisTimeStep, carCollisionPreviousTimeStep);
    }

    public void setRotationalVelocity(double rotationalVelocity) {
        this.angularVelocity = rotationalVelocity;
    }

    private double correctAngle (double angle) {
        while (angle < - Math.PI)
            angle += Math.PI * 2;
        while (angle > Math.PI)
            angle -= Math.PI * 2;
        return angle;
    }

    private void calculateSpeeds () {
        double velocityMagnitude = getVelocity ().mag ();
        directionOfMovement = Math.atan2 (velocity.y, velocity.x);
        // rotate
        double directionDifference = directionOfMovement - orientation;
        speed = Math.cos (directionDifference) * velocityMagnitude;
    }

    private Vector2d getDrivingForce (int command) {
        if (command == neutral || command == left || command == right){
            return new Vector2d (0, 0);
        }
        else if (command == forward || command == forwardleft || command == forwardright) {
            Vector2d force = new Vector2d (Math.cos (orientation), Math.sin (orientation));
            force.setMag (motorForceForward);
            return force;
        }
        else if (command == backward || command == backwardleft || command == backwardright) {
            Vector2d force = new Vector2d (-Math.cos (orientation), -Math.sin (orientation));
            force.setMag (motorForceBackward);
            return force;
        }
        System.out.println ("Car.getDrivingForce: Command " + command + " does not exist");
        return null;
    }

    private double getAVFromSteering (int command) {
        double angularMomentumFromSteering = 0;

        if (!((command == neutral) || (command == forward) || (command == backward))) {

            angularMomentumFromSteering = speed / turningRadius;
            if (command == left || command == forwardleft || command == backwardleft) {
                angularMomentumFromSteering *= -1;
            }
        }
        // Apply force from steering to angular momentum
        // Dependent on speed, limited by traction and orientation's
        // current deviation from direction of velocity
        double avFromSteering = angularMomentumFromSteering - angularVelocity;
        avFromSteering = Math.max (-maxLateralTyreTraction / 10,
                (Math.min (maxLateralTyreTraction / 10, avFromSteering)));
        return avFromSteering;
    }

    private Vector2d getLateralForce (double slipAngle) {
        double orthogonalDirection = (slipAngle > 0 ? -(Math.PI / 2) : (Math.PI / 2));
        Vector2d lateralForce = new Vector2d
                (Math.cos (orientation + orthogonalDirection),
                        Math.sin (orientation + orthogonalDirection));
        double lateralMagnitude = Math.min (speed, maxLateralTyreTraction);
        lateralForce.setMag (lateralMagnitude);
        return lateralForce;
    }

    // The following only applies to car-car collisions

    public Vector2d collisionWith (Vehicle other) {
        Vector2d otherPos = other.getPosition ();
        if (this.position.dist (otherPos) > 50) {
            // If we are far away from the other car, we are obviously not colliding
            return null;
        } else {
            // If we are close to the other vehicle, get its bounding polygon
            Polygon otherBounds = other.getPolygon ();
            // look if one of our corners are contained in the other polygon
            // in that case, return the coordinates of that corner as the point of collision
            // in a more advanced version, we might average the position of all corners that
            // are inside the bounds of the other vehicle, if more than one
            if (currentCorners == null) {
                createCurrentCorners (position, orientation);
            }
            for (int[] currentCorner : currentCorners) {
                if (otherBounds.contains(currentCorner[0], currentCorner[1])) {
                    return new Vector2d(currentCorner[0], currentCorner[1]);
                }
            }
        }
        return null;
    }

    public void impact (Vector2d otherPosition, Vector2d otherVelocity, double otherMass) {
        // Handle the impact of another vehicle with specified position, velocity and mass
        // Don't change anything in the other vehicle's state though, only this vehicle's
        // And only change velocity and angular velocity - this will happen before next() is called
        //double multiplier = otherMass / this.mass;
        // Currently, we are just taking the other vehicle's velocity and adopting it!
        // This assumes the other vehicle is doing the same with regards to our velocity.
        this.velocity = new Vector2d (otherVelocity.x, otherVelocity.y);
        // But this is not enough. We might still get problems with cars "hooking up to each others tails".
        // We should this add a few pixels movement away from the other car - if that position is not inside
        // a wall, of course...
        final int numberOfPixelsToMove = 1;
        Vector2d positionDifference = new Vector2d (position.x - otherPosition.x, position.y - otherPosition.y);
        double angleToOther = Math.atan2 (positionDifference.y, positionDifference.x);
        //System.out.println ("     car " + carID + " atan " + angleToOther);
        Vector2d potentialPosition = new Vector2d (position.x, position.y);
        potentialPosition.x += Math.cos (angleToOther) * numberOfPixelsToMove;
        potentialPosition.y += Math.sin (angleToOther) * numberOfPixelsToMove;
        // oh yes, we don't have walls in this version of the simulator
        position = potentialPosition;
        // now the angular velocity, depending on where the collision was, and the velocity difference...
        // how shall we do this? One way is to find out which corner the collision is closest to, and rotate
        // accordingly
        if (!carCollisionPreviousTimeStep) {
        int closestCorner = 0;
        int closestDistance = Integer.MAX_VALUE;
        createCurrentCorners (position, orientation);
        for (int i = 0; i < currentCorners.length; i++) {
            int thisDistance = (int) (new Vector2d (currentCorners[i][0], currentCorners[i][1]).dist (otherPosition));
            if (thisDistance < closestDistance) {
                closestDistance = thisDistance;
                closestCorner = i;
            }
        }
        // If the closest corner is front left or back right, deflect right
        if (closestCorner == 0 || closestCorner == 2) {
            angularVelocity -= ((otherVelocity.mag () + speed) / 2) / mass;
        }
        // else deflect left
        else {
            angularVelocity += ((otherVelocity.mag () + speed) / 2) / mass;
        }
        }
        carCollisionThisTimeStep = true;
    }

    private void createCurrentCorners (Vector2d newPosition, double orientation) {
        int x = (int) newPosition.x, y = (int) newPosition.y;
        currentCorners = new int[4][2];
        final int coslength = (int) (Math.cos (orientation) * 10);
        final int sinwidth = (int) (Math.sin (orientation) * 5);
        final int sinlength = (int) (Math.sin (orientation) * 10);
        final int coswidth = (int) (Math.cos (orientation) * 5);
        // front left
        currentCorners[0][0] = x + coslength + sinwidth;
        currentCorners[0][1] = y + sinlength - coswidth;
        // front right
        currentCorners[1][0] = x + coslength - sinwidth;
        currentCorners[1][1] = y + sinlength + coswidth;
        // back right
        currentCorners[2][0] = x - coslength - sinwidth;
        currentCorners[2][1] = y - sinlength + coswidth;
        // back left
        currentCorners[3][0] = x - coslength + sinwidth;
        currentCorners[3][1] = y - sinlength - coswidth;
    }    

    public Polygon getPolygon () {
        if (currentCorners == null) {
            createCurrentCorners (position, orientation);
        }
        if (polygon == null) {
            int[] xpoints = {currentCorners[0][0], currentCorners[1][0],
                    currentCorners[2][0], currentCorners[3][0]};
            int[] ypoints = {currentCorners[0][1], currentCorners[1][1],
                    currentCorners[2][1], currentCorners[3][1]};
            polygon = new Polygon (xpoints, ypoints, 4);
        }
        return polygon;
    }

    public boolean justCollided() {
        return carCollisionThisTimeStep;
    }

}
